
#ifndef DCF77_H
#define DCF77_H

#include "stdint.h"
#include "stdbool.h"
#include "math.h"

enum TDCF77_Error
{
  _ErrOk                  =  0, // kein Fehler
  _ErrDCT                 =  1, // fr "ADCSamplesPerPeriod" gibt es keinen Source-Code
  _ErrCarrier_RawOverflow =  2, // CarrierRawCntPerSecond zu klein
  _ErrCarrier_RawCnt      =  3, // CarrierRawCntPerSecond ist nicht Vielfaches von CarrierFineCntPerSecond
  _ErrTimedAdc            =  4, // fCPU zu gross bzw. ADCSamplesPerPeriod zu klein bzw. TIM2 braucht Prescaler
  _ErrAdcSampleTime       =  5, // fCPU zu klein bzw. ADCSamplesPerPeriod zu gro
  _ErrAdcBufSize          =  6, // PeriodsPerSample ist nicht vielfaches von IRQPerSample
  _ErrAdcBufOverrun       =  7, // fCPU zu klein oder der IRQ war zu lange blockiert (wenn selten drfte Err zurckgesetzt werden)
  _ErrAvr001Coeff         =  8, // "Avr001_SamplesPerPRNBit" passt nicht zu "SamplesPerPRNBit" - anpassen
};

enum TDCF77_Const
{
  fMHz                    = 32,    // STM32-CPU (zu klein->zu langsam fr ADC); Quarzgenauigkeit mit max. +-200 ppm erforderlich
  fCarrier                = 77500, // Senderkonstante
  DCFPeriodsPerPRNBit     = 120,   // Senderkonstante
  ADCRange                = 4096,  // STM32-Konstante
  SamplesPerPRNBit        = 6,     // SMPPerSecond sollte grade sein. deshalb fr SamplesPerPRNBit eigentlich nur 6 mglich
  ADCSamplesPerPeriod     = 12,    // untersttzt werden 3..12 Samples je Periode
  IRQPerSample            = 2,     // klein:viel RAM; gross:viel CPU-Last
  DCTAvr512CntPerSecond   = 32,    // zur Messung der mittleren Phase vom DCF-Signal (zu klein->ungenau; zu gro->Speicherverschwendung)
  CarrierRawCntPerSecond  = 125,   // grobe Messung der Trgerfrequenz (zu klein->ungenau; zu gro->Speicherverschwendung)
  CarrierRawSeconds       =   1,   // Medauer fr grobe Messung
  CarrierFineCntPerSecond =   5,   // genaue Messung der Trgerfrequenz (zu klein->ungenau; zu gro->Speicherverschwendung)
  CarrierFineSeconds      =  10,   // 2..10; Medauer (bei sehr schnellen Temperaturnderungen 2s und bei schwachem Signal eher 10s)
  CarrierDoCalc_dCycles   = 1,     // 0/1; wenn der Wert Carrier_dCycles_L8 gewnscht ist (oder FLASH-Bedarf egal)
  CarrierDoSynchronize    = 1,     // 0/1; wenn ADC bzw. TIM3 mit Trgerfrequenz synchron laufen soll (oder FLASH-Bedarf egal)
  PRNDriftFuse            = 0,     // 0/1; wenn Dauerbetrieb bei kleinem DCF77-Trgersignal (oder FLASH-Bedarf egal)

  // berechnet
  PeriodsPerSample  = DCFPeriodsPerPRNBit/SamplesPerPRNBit,
  Sample_MaxValue   = (int32_t)(256*ADCRange/2*PeriodsPerSample*ADCSamplesPerPeriod*2/M_PI), // Uss=VCC; Rechteck
  AdcCntPerSample   = PeriodsPerSample*ADCSamplesPerPeriod,
  AdcBufCnt         = 2*AdcCntPerSample/IRQPerSample,
  CarrierRawArrCnt  = (int32_t)(CarrierRawCntPerSecond *(CarrierRawSeconds +0.5)), // +0.5s fr Bearbeitung in Carrier_GetResult()
  CarrierFineArrCnt = (int32_t)(CarrierFineCntPerSecond*(CarrierFineSeconds+1.0)), // +1.0s fr Bearbeitung in Carrier_GetResult()
  DCTAvr512Cnt      = DCTAvr512CntPerSecond,
  SMPPerSecond      = (fCarrier*SamplesPerPRNBit+DCFPeriodsPerPRNBit-1)/DCFPeriodsPerPRNBit,

  // Debug
  DbgSample_Cnt = 0*2048,
};

#define Avr001_SamplesPerPRNBit 6

// hier sind die Ergebnisse (werden laufend vom IRQ aktualisiert)
typedef struct
{
  int8_t   Error;                 // Fehler vom Typ TDCF77_Error (meist Konfigurationsfehler)
  int32_t  SMPLoop;               // aktuelle "Modulzeit" in ganzen Sekunden
  int32_t  SMPOfs;                // aktuelle "Modulzeit" innerhalb der Sekunde (0..SMPPerSecond-1)
  int32_t  Carrier_SNR_Raw_dB;    // (ein) Sender gefunden, wenn dB-Werte hufiger positiv als negativ sind
  int32_t  Carrier_SNR_Fine_dB_E1;// ab ca. 3.0 db (d.h. Carrier_SNR_Fine_dB_E1>=30) ist genauer Abgleich erfolgt
  int32_t  Carrier_dCycles_L8;    // Abweichung der "Modulzeit" zur "Atomzeit" in 256tel-Trgerperioden; s.u. "Stoppuhr"
  int32_t  Carrier_Synchronized;  // falls !=0: "Modulzeit" seit wann synchron zum Trger (s.o. "CarrierDoSynchronize")
  int32_t  Carrier_dfHSE_E9;      // ermittelte STM32-Quarzabweichung in 0.001 ppm (nur gltig wenn Sender gefunden)
  int32_t  Signal_Ueff_uV;        // Trgersignalpegel (fCarrier+-0.5Hz) am ADC-Eingang (nur sinnvolle Werte wenn Sender gefunden)
  int32_t  DCF77_SMPOfs;          // Zeitpunkt vom DCF77-Sekundenbeginn (s.u. "Beziehung zwischen Modulzeit und STM32-Peripherie")
  int32_t  DCF77_RstCnt;          // Reset (da BitPos_SMPOfs gesprungen ist)
  int32_t  DCF77_AvrAbsHitVal;    // 0..256; mittlerer Absolutwert vom HitVal; reagiert trge; quasi "Signalqualitt"; bis 9 Rauschen
  int32_t  DCF77_HitVal;          // -256..+256; Trefferwert vom DCF-Bit; "Vorzeichen=Bitwert"; "Betrag=Qualitt"
  int32_t  DCF77_BitVal;          // 0/1; das DCF-Bit
  int32_t  DCF77_BitCnt;          // wieviele DCF-Bits gehren zur aktuellen Sequenz (wann war letzter Reset ntig)
  int32_t  DCF77_TotalBitCnt;     // Gesamtbitzahl (zur Erkennung einer nderung/Aktualisierung)
} volatile TDCF77State;

// (... und ab hier sind nur interne Strukturen:)

typedef struct
{
  int16_t ValArr[AdcBufCnt]; // Puffer ADC-Werte
  uint8_t TimArr[AdcBufCnt]; // ADC-Sample-Zeitpunktabstnde
  int32_t w;
  int32_t n;
  int32_t a0;
  int32_t a1;
  int32_t d_ppm_L10;
} TTimedAdc;

typedef struct
{
  int32_t  RawSubCur;
  int32_t  RawPhaCur;
  uint8_t  RawPhaArr[CarrierRawArrCnt];
  int32_t  PrevC;
  int32_t  PrevS;
  int32_t  SumC;
  int32_t  SumS;
  int32_t  FineSumC;
  int32_t  FineSumS;
  int32_t  FineSubCur;
  int32_t  FinePhaCur;
  uint8_t  FinePhaArr[CarrierFineArrCnt];
  int32_t  FinePpmArr[CarrierFineArrCnt];
  int32_t  FinePhaCnt;
  int32_t  RawP;
  int32_t  RawN;
  int32_t  RawN1;
  int32_t  RawNE;
  bool     AlwaysSmooth;
  int32_t* pd_ppm_L10;   // Zeiger auf gemeinsamen Wert mit TTimedAdc
  int16_t  Cur_ppm_L0;   // ganzzahliger Teil der ppm-Abweichung
  int16_t  Cur_ppm_L10;  // aktueller Rest in 1/1024 ppm
  int16_t  Next_ppm_L10; // anzuwendender (zum Sync mit FinePpmArr[]) Rest
  int32_t  Next_ppm_L10_Avr;
  int32_t  dCycles_L8;
  int32_t  Sync_ppm_L10_Avr;
  int32_t  Sync_FinePhaOfs;
} TCarrier;

typedef struct
{
  int32_t AvrC; // DCT-Cos-Koeffizient ber grob eine Sekunde
  int32_t AvrS; // DCT-Sin-Koeffizient ber grob eine Sekunde
  int32_t ArrC[DCTAvr512Cnt];
  int32_t ArrS[DCTAvr512Cnt];
  int32_t Cnt1;
  int32_t Cnt2;
  int32_t Ofs;
  int32_t SumC;
  int32_t SumS;
  int32_t PrevS;
  int32_t PrevC;
  int32_t ResetCnt;
} TDCTAvr512;

// - DCTAvr001_Coeff[]
//   Fr andere SamplesPerPRNBit kann eine Rechteck-Fensterfunktion genommen
//   werden: Ein TDCTAvr001_Coeff[]-Feld mit SamplesPerPRNBit Werten, alle 255.
#if Avr001_SamplesPerPRNBit==3
  // Rechteck-Fensterfunktion
  static const uint8_t DCTAvr001_Coeff[3] =
  {
    255,
    255,
    255,
  };
#endif
#if Avr001_SamplesPerPRNBit==4
  // Rechteck-Fensterfunktion
  static const uint8_t DCTAvr001_Coeff[4] =
  {
    255,
    255,
    255,
    255,
  };
#endif
#if Avr001_SamplesPerPRNBit==6
  // Blackman-Harris-Fensterfunktion
  static const uint8_t DCTAvr001_Coeff[14-2] =
  {
      1,
     10,
     42,
    106,
    187,
    246,
    246,
    187,
    106,
     42,
     10,
      1,
  };
#endif

typedef struct
{
  int32_t PrevC;
  int32_t PrevS;
  int32_t ArrC[sizeof(DCTAvr001_Coeff)/sizeof(*DCTAvr001_Coeff)];
  int32_t ArrS[sizeof(DCTAvr001_Coeff)/sizeof(*DCTAvr001_Coeff)];
  int32_t Ofs;
} TDCTAvr001;

typedef struct
{
  uint32_t Arr[SamplesPerPRNBit][512/32]; // Ringpuffer
  int32_t  P; // Phase im PRNBit (0..SamplesPerPRNBit-1)
  int32_t  I; // Nr. vom PRNBit (0..512)
} TPRNBitTst;

enum TPRNBitOfs_Const
{
  PRNBitOfs_HitArrCnt  = SamplesPerPRNBit+1,
  PRNBitOfs_ProbeCnt   = 4,
  PRNBitOfs_FocusTrack = 650,
  PRNBitOfs_ProbeInit  = 8, // falls zu klein wird Fokus bei Minimumsignal nicht gehalten; zu gro -> wird nie gefunden
  PRNBitOfs_ProbeLimit = 500000*PRNBitOfs_HitArrCnt/2,
  PRNBitOfs_dFocusPos  = SamplesPerPRNBit,
};

typedef struct
{
  int32_t MaxVal;
  int32_t MaxPos;
  int32_t HitOfs;
  int32_t HitAvr;
  int32_t FocusPos;
  int8_t  HitArr[PRNBitOfs_HitArrCnt];
  int32_t ProbeVal[PRNBitOfs_ProbeCnt];
  int16_t ProbePos[PRNBitOfs_ProbeCnt];
} TPRNBitOfs;

enum TDCF77Control_Const
{
  DCF77Control_dpAvr   = SamplesPerPRNBit/3,
  DCF77Control_sAvrTau = 10, // Zeitkonstante in s fr DCF77_AvrVal
};

typedef struct
{
  TTimedAdc  TimedAdc;
  TCarrier   Carrier;
  TDCTAvr512 DCTAvr512;
  TDCTAvr001 DCTAvr001;
  TPRNBitTst PRNBitTst;
  TPRNBitOfs PRNBitOfs;
  int32_t  SMPOfs;
  int32_t  SMPOfsFrac;
  int32_t  DCF_SMPOfs_Next;
  int32_t  DCF_SMPOfs;
  int32_t  DCF_AvrVal;
  int32_t  DCF_HitVal;
  int32_t  DCF_HitSum;
  int32_t  CurC;
  int32_t  CurS;
  uint32_t Rand;
  int32_t  AdcBufOfs;
  int32_t  IRQCur;
  bool     OscHop;
  int16_t  OSC_Drift;
  int16_t  OSC_Ref_e7;
  int8_t   OSC_Time;
} TDCF77Control;

void DCF77_Init(TDCF77Control* Control, TDCF77State* State, int32_t ppm);
  // DCF77-Modul einschalten
  //  - HSE-Frequenz zuvor (selber) entsprechend "fMHz" (s.o.) einstellen
  //  - Parameter
  //    - Control : nur bereitstellen (es gibt nichts zu sehen oder zu ndern)
  //    - State   : Diese Struktur wird vom DCF77-Modul (TIM3_IRQHandler())
  //                laufend aktualisiert. Hier erhlt man smtliche
  //                Informationen. Die Inhalt darf nicht verndert werden.
  //    - ppm     : Startwert fr STM32-Quarzabweichung (normalerweise 0)
  //                - Wert kann beim vorherigen Durchlauf ermittelt werden
  //                  ("ppm = DCF77State.Carrier_dfHSE_E9/1000")
  //                - bei ppm==0
  //                  - vollstndig automatische Erkennung
  //                  - dauert etwa 2 Sekunden
  //                    (whrend Erkennung keine DCF-Bits)
  //                - bei ppm auf 5 PPM genau
  //                  - teilweise automatische Erkennung
  //                  - dauert etwa 1 Sekunden
  //                - bei ppm auf 1 PPM genau
  //                  - keine Verzgerung durch Messung
  //                    der STM32-Quarzabweichung

void DCF77_Deinit(TDCF77Control* s);
  // DCF77-Modul abschalten

void DCF77_Calib_fADC(TDCF77Control* s);
  // fADC-Kalibrierung - wichtig!
  // - in der Regel 1 x je Sekunde aufrufen
  // - sinnvollerweise dann, wenn sich DCF77_TotalBitCnt ndert (die
  //   langwierige Berechnung fllt dann hauptschlich in die
  //   DCF-Bit-Pause (0.2s) und verursacht dann weniger ADC-Strungen)
  // - kann je nach Konfiguration recht lange dauern (1ms..250ms)
  //   - Z.B. 150ms bei fMHz==8, SamplesPerPRNBit=3, ADCSamplesPerPeriod=3.

void DCF77_Suspend(TDCF77Control* s);
  // (sehr spezielle Funktion)
  // Falls einige DCF-Bits nicht bentigt werden und Strom gespart werden
  //  soll, kann die Verarbeitung angehalten werden.
  // - das Anhalten sollte nicht whrend der bertragung eines DCF-Bits
  //   erfolgen, da im nchsten DCF77_BitVal sonst zwei DCF-Bits anteilig
  //   vermischt sind (die bertragung eines Bits beginnt 0.2s nach
  //   DCF-Sekundenbeginn und dauert 0.8s)
  // - sinnvoll wre ein Anhalten nach nderung von DCF77_TotalBitCnt
  //   (ist der Zeitpunkt vom DCF-Bit-Ende)
  // - DCF77_Resume() muss nach einer Ganzzahl einer Sekunde auf 250s genau
  //   aufgerufen werden; die Zeitspanne beginnt nach(!) dem Aufruf von
  //   DCF77_Suspend()
  // - vor DCF77_Resume() muss DCF77_ResumeInit() aufgerufen werden
  // - eine driftende Quarzfrequenz wird erst drei Sekunden nach einem
  //   DCF77_Resume() ausgeglichen; gegebenenfalls entweder mind. drei Bits
  //   lesen (aktiv bleiben) oder nach nchsten Resume drei DCF-Bits verwerfen
  // - bei groen Suspendzeiten knnte die Drift von DCF77_SMPOfs verwendet
  //   werden um STM32-Quarzabweichung oder Zeitspannen alternativ zu
  //   bestimmen
  // - der erste DCF77_BitVal nach dem Resume knnte einen schlecheren Wert
  //   haben; bei schwachen Signal evtl. relevant

void DCF77_ResumeInit(TDCF77Control* s);
  // s.o. DCF77_Suspend()
  // - nach DCF77_Suspend() und vor DCF77_Resume() aufrufen

void DCF77_Resume(TDCF77Control* s);
  // s.o. DCF77_Suspend()
  // - Aufruf von DCF77_ResumeInit() nicht vergessen

#endif

  //
  // ### DCF77-Modul ###
  //
  // Das Modul wertet direkt DCF77-Antennendaten aus. Ein 77kHz-Verstrker
  // (100x..1000x) ist aber meist ntig. Anders als bei anderen Empfngern
  // wird statt der Amplituden- die Phasenmodulation ausgewertet.
  // Das Modul gibt die DCF-Datenbits "DCF77_BitVal" zurck aus dem sich Datum
  // und Uhrzeit bestimmen lassen. Zu jedem DCF-Datenbitwert gibt es zustzlich
  // eine Information ber die Fehlerwahrscheinlichkeit "DCF77_HitVal".
  // Die absolute Zeit ist auf etwa 1ms genau. Zeitspannen sind mit etwa 1s
  // Genauigkeit messbar ("Carrier_dCycles_L8").
  // Der Code ist fr den STM32F103C8T6 geschrieben. Der Code wre aber auch
  // auf kleineren Controllern (z.B. STM32F030F4P6) lauffhig, wenn die
  // Peripheriezugriffe angepasst werden.
  //
  // Dank an Sean Eron Anderson fr die schne Zusammenfassung von
  // "Bit Twiddling Hacks"
  // ("https://graphics.stanford.edu/~seander/bithacks.html").
  //
  // Dank auch an die "d.velop AG" fr die "freundliche" finanzielle
  // Untersttzung.
  //
  // Copyright: Public Domain
  //
  // Dislaimer: Keine Haftung fr Schden.
  //
  // C.B., 2021
  //
  // - Nachteile gegenber normalen Empfngern
  //   - Der Stromverbrauch ist mit 15mA enorm (1000x). Ein Betrieb mit
  //     Knopfzellen ist nicht mglich.
  //   - Der Controller strt den Empfang. Eine kompakte Bauweise geht nur
  //     bedingt (s. "STM32-Strungen mindern" in "dcf77.h").
  //   - Normalerweise unkritische Stromspitzen auf den GND-/VCC-Leitungen
  //     sind sehr strend. (s. "Verwendung" in "dcf77.h")
  //   - Eventuell mssen whrend des Empfangs andere Ttigkeiten
  //     (Display-Ansteuerung, ...) eingestellt werden (s.u. "Verwendung").
  //   - Das Programm (inkl. Dekoder) belegt mit etwa 9 kB einen wesentlichen
  //     Teil vom Flash.
  //   - bla
  //     - Fremdsoftware erzeugt Abhngigkeiten.
  //     - Fremdsoftware ist selten Plug&Play.
  // - Verwendung
  //   - Hardware-Todo
  //     - DCF77-Analog-Signal an GPIO-PA7 anlegen
  //       - ca. 0.1mV bis 1000mV
  //       - fr eine 1cm Ferritspule wre ca. eine 500x Vorverstrkung ntig
  //       - Frequenzen ber 77500Hz per RC-Tiefpass nahe am GPIO-PA7
  //         dmpfen
  //       - Mindestbandbreite 1kHz
  //     - Antennenausrichtung beachten (Ferritachse waagerecht zum Boden
  //       und senkrecht zu Frankfurt-Mainflingen)
  //     - Strquellen vom Schwingkreis fernhalten (TFT:1Meter, STM32 10cm,
  //       ...)
  //     - Schwingkreis
  //       - Temperaturstabilitt vom Kondensator prfen
  //         (z.B. per Fn whrend der Resonanzfrequenzmessung)
  //       - Resonanzfrequenz nochmals einige Stunden nach dem Anlten des
  //         Kondensators prfen
  //       - mind. 1kHz Bandbreite
  //       - je hher die Bandbreite desto geringer die Signalstrke
  //       - auch Kondensatorverluste erhhen die Bandbreite
  //     - jhrliche Signalschwankungen einplanen (s. Bild 16 in
  //       "PTB-Mitteilungen 114 (2004), Heft 4")
  //     - Verbindung Schwingkreis zum Verstrker kurz halten.
  //     - Verstrker-VCC muss unempfindlich gegenber Strungen sein oder
  //       dazu gemacht werden (z.B. RC-Tiefpass).
  //       - zum Stromsparen knnte der Verstrker ber ein GPIO-Pin
  //         abschaltbar mit Strom versorgt werden
  //     - Stromspitzen auf der VCC-Zuleitung unterbinden, da sonst andere
  //       damit verbundene Teile (z.B. Batterien) zu Strquellen werden
  //       knnten. (Stromspitzen entstehen z.B. bereits dadurch, dass eine
  //       Schaltung mit Elko ber ein Kabel an den STM32 angeschlossen ist.)
  //       - z.B. per Dmpfungsperle (besser hhere Induktivitt)
  //       - oder ein ber Widerstand versorgter "persnlicher" Spannungsregler
  //         mit Elko in Controller-Nhe
  //     - Prfen ob eine abgeschirmte oder verdrillte Zuleitung vom Verstrker
  //       zum STM-Board vorteilhaft ist.
  //     - Prfen inwiefern interne Teile (Batterie, Spannungsregler, LCD, ...)
  //       Strungen verursachen
  //       - z.B. durch Abklemmen smtliche Teile und der Stromversorgung ber
  //         eine passende Batterie
  //       - Leitungen mit Stromnderungen oder deren Oberwellen im Bereich
  //         von 77500Hz+-650Hz sind dabei besonders strend
  //     - Bei einer elektrischen Verbindung zu Strquellen (Schaltnetzteil,
  //       Laptop, ...) knnte eine Erdung helfen Strungen zu verringern.
  //     - Ein Sendersignal kann auch ber das Stromnetz oder die Erdung
  //       aufgenommen werden. Das kann zu seltsamen Effekten fhren:
  //       - der Empfang ist auch ohne Schwingkreis mglich
  //       - die optimale Antennenausrichtung passt nicht zur Senderrichtung
  //       - eine Drehung um 180 hat Einfluss auf die Signalstrke
  //       Falls beim Debuggen/Messen strend, dann knnten die Daten ber
  //       einen Optokoppler oder ADuMxxxx bertragen werden. Zur
  //       Stromversorgung mssen Batterien verwendet werden. Der DC-DC-Wandler
  //       eines USB-Isolators entkoppeln vermutlich nicht gengend (>10pF).
  //     - ###Tipp###
  //       Die Auswirkung (Signalqualittsnderung) einer Software-/Hardware-
  //       nderung lsst sich schnell und sicher anhand des
  //       "DCF77_AvrAbsHitVal"-Parameters beurteilen. Alternativ auch mit
  //       "DCF77_HitVal" (reagiert sofort aber dafr schwankend).
  //   - Software-Todo
  //     - Die CPU-Frequenz (selbstverstndlich HSE und nicht HSI) selber
  //       (entsprechend "fMHz" s.o. "TDCF77_Const") "einstellen".
  //     - einmalig DCF77_Init() aufrufen
  //     - "DCF77State.Error" prfen ob ein Problem vorliegt (die meisten
  //       Fehler sind statisch, manche treten aber erst spter auf!)
  //     - das Ergebnis aus den DCF77State-Daten auslesen - wichtig sind:
  //       DCF77_HitVal : der Trefferwert vom DCF-Bit
  //                      <0 : DCF-Bit ist 0
  //                      >0 : DCF-Bit ist 1
  //                      Falls nur das DCF-Bit bentigt wird, siehe
  //                      "DCF77_BitVal".
  //       DCF77_BitVal : das DCF-Bit
  //                      (Meist gilt DCF77_BitVal == (DCF77_HitVal>=0)". Falls
  //                      nicht, ist "DCF77_BitVal" der wahrscheinlichere
  //                      Wert.)
  //                      Achtung: Die Kodierung beim phasenmodulierten Signal
  //                      weicht etwas von der beim amplitudenmod. Signal ab.
  //                      Siehe Abschnitt "PRN-modulierten Signal" in
  //                      "PTB-Mitteilungen 114 (2004), Heft 4".
  //       DCF77_BitCnt : Bei nderung um eins gibt es einen neuen
  //                      DCF77_*-Datensatz. Bei nderungen weder 0 oder 1 ist
  //                      die Bitfolge gestrt. Die Auswertung der
  //                      DCF77_BitVal-Werte sollte neu begonnen werden.
  //     - Wenn der STM32 Ausgnge schalten muss (LCD, USART, ...), dann sollte
  //       das in den ersten 200ms nach einer nderung von DCF77_TotalBitCnt
  //       geschehen. In diesem Intervall werden keine PRN-Bits gesendet und
  //       es entsteht deswegen kein Qualittsverlust.
  //       Dannach sollte die Strungszeit mglichst kurz (z.B. durch hohe
  //       bertragungsrate) sein um mglichst wenige Bits (insgesamt gibt es
  //       512 in den verbleibenden Interval von 800ms) zu zerstren.
  //     - Synchronisierung beim Zugriff auf DCF77State-Daten
  //       Die Struktur wird vom DCF77-Modul (TIM3_IRQHandler()) laufend
  //       aktualisiert. Eine Aktualisierung lsst sich zuverlssig am SMPOfs
  //       erkennen. Wenn sich vor und nach dem Lesen von Elementen der SMPOfs
  //       nicht gendert hat, dann passt das gelesene zueinander (anderfalls
  //       nachmal lesen). Wirklich wichtig ist dies eigentlich nur beim Lesen
  //       vom SMPLoop/SMPOfs-Paar. Ein Sperren vom IRQ ist jedenfalls nie
  //       ntig.
  //     - "Vertrglichkeit" mit anderen (nicht DCF77-Modul-) IRQs:
  //       Der IRQ (TIM3_IRQHandler()) darf nur sehr kurz (z.B. von
  //       __disable_irq() oder anderen IRQs) blockiert sein. Notfalls
  //       AdcBufCnt erhhen (von "2*AdcCntPerSample/IRQPerSample" auf
  //       "x*AdcCntPerSample/IRQPerSample"). Zu lange Blockaden verschlechern
  //       die Qualitt und sind am "DCF77State.Error"-wert erkennbar.
  //     - "__WFI();" beim Warten auf DCF77-Bits verwenden
  //       Die Strungen werden geringer und die Stromaufnahme sinkt.
  //     - STM32-Strungen mindern
  //       Falls der Abstand zur Ferritspule nicht vergrert werden kann,
  //       dann knnen durch eine andere Konfiguration die Strungen vom
  //       STM32 gemindert werden (alle 4 nderungen durchfhren):
  //       - TDCF77_Const::fMHz                von 32 nach 8
  //       - TDCF77_Const::SamplesPerPRNBit    von  6 nach 3
  //       - TDCF77_Const::ADCSamplesPerPeriod von 12 nach 3
  //       - "#define Avr001_SamplesPerPRNBit" von  6 nach 3
  //       Auerdem probieren, ob ein Verzicht von "__WFI();" besser ist (mehr
  //       Strungen, dafr gleichmiger)
  //   - Beziehung zwischen Modulzeit und STM32-Peripherie
  //     Mit TIM3 kann die "Modulzeit" (SMPLoop, SMPOfs) exakt einem ADC-
  //     Samplezeitpunkt eines Samples zugeordnet werden:
  //     Der Sample&Hold des ersten ADC-Samples eines Sample N ist genau
  //     dann, wenn TIM3->CNT auf
  //       N*ADCSamplesPerPeriod*DCFPeriodsPerPRNBit/SamplesPerPRNBit
  //     springt (mit N = SMPLoop*SMPPerSecond + SMPOfs).
  //     Mit den noch freien Kanlen (TIM3_CCR2 bis TIM3_CCR4) knnen
  //     Hardwarevorgnge zeitlich genau (Sekundenimpuls am STM32-GPIO-Pin,
  //     ...) gemessen / gesteuert werden.
  //   - Zeitspannenmessung ("Stoppuhr")
  //     TIM3 zhlt ADC-Samples:
  //           nSamples = (fCarrier+fError)*ADCSamplesPerPeriod
  //     Aus "DCF77State.Carrier_dCycles_L8" kann das jeweils (fast) aktuelle
  //     (Mittelwert der letzten 10s) fError berechnet werden:
  //            fError = DCF77State.Carrier_dCycles_L8/256
  //     "Atomgenaue" Zeitdifferenzen knnen also ber die TIM3-Taktnderung
  //     (alle berlufe bercksichtigen!) zusammen mit der nderung von
  //     Carrier_dCycles_L8 ber die Mezeit bestimmt werden. Der Fehler
  //     liegt bei +-1s zuzglich sonstigen Fehlern (+-10s?; s.
  //     "PTB-Mitteilungen 114 (2004), Heft 4"). Es ist (mir) unklar ob es
  //     bei groer Senderentfernung und bei Dunkelheit (berlagerung der
  //     Raumwelle) mglicherweise zu einem unerkannten (da kaschiert) Versatz
  //     von Schwingungen kommen knnte.
  //     - Der Schalter "CarrierDoCalc_dCycles" muss auf 1 gesetzt sein. (Der
  //       Schalterwert "CarrierDoSynchronize" ist egal.)
  //     - Achtung: Der Wert Carrier_dCycles_L8 ist nicht aktuell, sondern
  //       ber CarrierFineSeconds gemittelt (also auch veraltet)!
  //     - Fr ein genaues Carrier_dCycles_L8 muss die Temperatur ausgeglichen
  //       sein. Falls nicht mglich oder Wartezeit zu gross, kann
  //       CarrierFineSeconds von 10 auf 2 verringert werden. Dafr ist aber
  //       ein besserer Empfang ntig (DCF77_AvrAbsHitVal>60).
  //     - Voraussetzung ist, dass der Empfang ununterbrochen ber die gesamte
  //       Mezeit ausreichend ist:
  //         Carrier_SNR_Fine_dB_E1 > 20
  //         DCF77_BitCnt           > 10
  //         DCF77_AvrAbsHitVal     > 20
  //         DCF77_RstCnt           gleichbleibend
  //         Antennenausrichtung    gleichbleibend - Wichtig!
  //   - Stromsparen
  //     Geht etwas (max. 25%) ber Verringerung von fCPU und Rechenbedarf.
  //     Dazu ADCSamplesPerPeriod bis auf 3 und notfalls ("DCF77.c" untersttzt
  //     das nicht richtig) auch SamplesPerPRNBit auf 3 verkleinert. Die
  //     Qualitt sinkt dadurch.
  //   - Fehlerursachen
  //     1. falls "DCF77State.Error" ungleich 0
  //        -> Fehler: falsch konfiguriert
  //     2. falls "DCF77State.Carrier_SNR_Raw_dB"-Wert fters negativ als
  //        positiv
  //        -> Fehler: das DCF77-Signal an GPIO-PA7 fehlt oder die CPU-Frequenz
  //           wurde ("selber machen") nicht auf "TDCF77_Const.fMHz" (max.
  //           +-200 ppm Abweichung) gesetzt
  //     3. falls "DCF77State.DCF77_BitCnt" immer auf 0 zurck springt
  //        -> Bei sehr schwachem Signal oder starken Strungen kann es bis
  //           zum stabilen Empfang ber 60 Sekunden lang zu Abbrchen kommen.
  //     4. falls Carrier_dCycles_L8 trotz CarrierDoSynchronize kaum abnimmt
  //        oder sogar ansteigt
  //        -> kein Fehler: HSE driftet wegen Temperaturnderung
  //           (Umgebungstemperatur, Eigenerwrmung) und die Regelung hat
  //           absichtlich eine langsame Ausgleichszeit (Tau ca. 20s; und Delta
  //           max. 2/s) um den PRNBit-Dekoder nur wenig zu stren.
  //     5. falls Carrier_dCycles_L8 stark schwankt
  //        -> starke Strungen oder Empfangssituation (nderung der
  //           Reflexionsanteile der Schichten der Ionosphre)
  //
  // - verwendete STM32F103-Hardware
  //   - rund 3k RAM (8 Byte statisch; 2296 Byte dynamisch; 292 Byte Stapel)
  //   - rund 5k FLASH (4900 Bytes)
  //   - TIM2
  //   - TIM3
  //   - DMA1_Channel1
  //   - DMA1_Channel2
  //   - ADC1
  //   - GPIO-PA7
  //
  // - Bitfehlerquote im Abhngigkeit vom DCF77_AvrAbsHitVal
  //   Die Werte gelten bei gleichmigem Strfeld (natrliches Rauschen,
  //   Elektronik, ...) und nicht bei nderungen (Gert Ein-/Abschalten).
  //
  //   AvrAbsHitVal Fehlerquote
  //   ------------ -----------
  //         9          50 %
  //        10          30 %
  //        11          25 %
  //        12          19 %
  //        13          15 %
  //        14          13 %
  //        15          11 %
  //        16         7.9 %
  //        17         6.7 %
  //        18         5.6 %
  //        19         4.7 %
  //        20         3.9 %
  //        21         3.2 %
  //        22         2.6 %
  //        23         2.1 %
  //        24         1.7 %
  //        25         1.4 %
  //        26         1.1 %
  //        27        0.85 %
  //        28        0.67 %
  //        29        0.52 %
  //        30        0.40 %
  //        31        0.31 %
  //        32        0.23 %
  //        33        0.18 %
  //        34        0.13 %
  //        35        0.10 %
  //        36        0.07 %
  //        37        0.05 %
  //        38        0.04 %
  //        39        0.03 %
  //
  // - Fehlerwahrscheinlichkeit eines Bits im Abhngigkeit vom DCF77_HitVal
  //   Fr jede Signalstrke und Rausch-/Strquelle gibt es eine eigene
  //   Verteilung. Die Werte hier passen fr den Fall, wenn ein normales
  //   Sendesignal (kein sehr niedriges/hohes Signal) durch eine groe Strung
  //   berlagert wird. Also in etwa die Wortcase-Fehlerwahrscheinlichkeit fr
  //   das momentane Bit (DCF77_BitVal).
  //   Die Tabelle zeigt Mittelwerte, Fr _einen_ DCF77_HitVal sind die vielen
  //   Kommastellen natrlich Unsinn.
  //   Beispiel:
  //   Wenn jeden Tag 60 Bits mit einem DCF77_HitVal=56 gelesen werden wrden,
  //   dann wre im Mittel alle 100 Jahre eines der Bits falsch.
  //   (100Jahre=1/0.000000423338137591/60/365)
  //
  //   HitVal Fehlerwahrscheinlichkeit
  //   ------  -----------------------------------
  //      ...
  //       10  0.200552715214199180  2.005527e-001
  //       11  0.176689182679008040  1.766892e-001
  //       12  0.154702557197146540  1.547026e-001
  //       13  0.134602843603653770  1.346028e-001
  //       14  0.116371133132232810  1.163711e-001
  //       15  0.099962593707953948  9.996259e-002
  //       16  0.085309949646420788  8.530995e-002
  //       17  0.072327276047782954  7.232728e-002
  //       18  0.060913936620409039  6.091394e-002
  //       19  0.050958505514050040  5.095851e-002
  //       20  0.042342532411092065  4.234253e-002
  //       21  0.034944033768334673  3.494403e-002
  //       22  0.028640619762230906  2.864062e-002
  //       23  0.023312194253474125  2.331219e-002
  //       24  0.018843192213871662  1.884319e-002
  //       25  0.015124344088059611  1.512434e-002
  //       26  0.012053978375645393  1.205398e-002
  //       27  0.009538891568667787  9.538892e-003
  //       28  0.007494828085965141  7.494828e-003
  //       29  0.005846621967870400  5.846622e-003
  //       30  0.004528057073394608  4.528057e-003
  //       31  0.003481503817988997  3.481504e-003
  //       32  0.002657388710944857  2.657389e-003
  //       33  0.002013548783566623  2.013549e-003
  //       34  0.001514517144353044  1.514517e-003
  //       35  0.001130779021785360  1.130779e-003
  //       36  0.000838030350960529  8.380304e-004
  //       37  0.000616463719959544  6.164637e-004
  //       38  0.000450099696682354  4.500997e-004
  //       39  0.000326175475261590  3.261755e-004
  //       40  0.000234597576042178  2.345976e-004
  //       41  0.000167461075600920  1.674611e-004
  //       42  0.000118634529825460  1.186345e-004
  //       43  0.000083407323980748  8.340732e-005
  //       44  0.000058194541202259  5.819454e-005
  //       45  0.000040293465429532  4.029347e-005
  //       46  0.000027685398772329  2.768540e-005
  //       47  0.000018876451538322  1.887645e-005
  //       48  0.000012771240584061  1.277124e-005
  //       49  0.000008573908053006  8.573908e-006
  //       50  0.000005711464884286  5.711465e-006
  //       51  0.000003775106270153  3.775106e-006
  //       52  0.000002475790717998  2.475791e-006
  //       53  0.000001610986535557  1.610987e-006
  //       54  0.000001040047852004  1.040048e-006
  //       55  0.000000666175101161  6.661751e-007
  //       56  0.000000423338137591  4.233381e-007
  //       57  0.000000266895093753  2.668951e-007
  //       58  0.000000166931487466  1.669315e-007
  //       59  0.000000103578756094  1.035788e-007
  //       60  0.000000063757039231  6.375704e-008
  //       61  0.000000038931475237  3.893148e-008
  //       62  0.000000023581915165  2.358192e-008
  //       63  0.000000014169449084  1.416945e-008
  //       64  0.000000008445253410  8.445253e-009
  //       65  0.000000004992847895  4.992848e-009
  //       66  0.000000002927857680  2.927858e-009
  //       67  0.000000001702972179  1.702972e-009
  //       68  0.000000000982451297  9.824513e-010
  //       69  0.000000000562147449  5.621474e-010
  //       70  0.000000000319017838  3.190178e-010
  //       71  0.000000000179553921  1.795539e-010
  //       72  0.000000000100225821  1.002258e-010
  //       73  0.000000000055482837  5.548284e-011
  //       74  0.000000000030459406  3.045941e-011
  //       75  0.000000000016582776  1.658278e-011
  //       76  0.000000000008952725  8.952725e-012
  //       77  0.000000000004792969  4.792969e-012
  //       78  0.000000000002544452  2.544452e-012
  //       79  0.000000000001339409  1.339409e-012
  //       80  0.000000000000699117  6.991168e-013
  //       81  0.000000000000361820  3.618203e-013
  //       82  0.000000000000185665  1.856654e-013
  //       83  0.000000000000094461  9.446093e-014
  //       84  0.000000000000047648  4.764802e-014
  //       85  0.000000000000023829  2.382852e-014
  //       86  0.000000000000011814  1.181399e-014
  //       87  0.000000000000005807  5.806720e-015
  //       88  0.000000000000002829  2.829358e-015
  //       89  0.000000000000001367  1.366642e-015
  //       90  0.000000000000000654  6.543635e-016
  //       91  0.000000000000000311  3.105757e-016
  //       92  0.000000000000000146  1.461123e-016
  //       93  0.000000000000000068  6.813404e-017
  //       94  0.000000000000000031  3.149094e-017
  //       95  0.000000000000000014  1.442573e-017
  //       96  0.000000000000000007  6.549475e-018
  //       97  0.000000000000000003  2.946984e-018
  //       98  0.000000000000000001  1.314127e-018
  //       99  0.000000000000000001  5.807253e-019
  //      ...
  //      256  0.000000000000000000  7.458341e-155
