
#include "DCF77DecBit.h"
#include "stdbool.h"


//############################################################## IsOddParity
static bool IsOddParity(uint32_t v)
{
  // "Compute parity in parallel "
  // ("https://graphics.stanford.edu/~seander/bithacks.html")
  v ^= v >> 16;
  v ^= v >> 8;
  v ^= v >> 4;
  v &= 0xf;
  return (0x6996 >> v) & 1;
}

//############################################################## DCF77DecBit_ProcessVal
void DCF77DecBit_ProcessVal(TDCF77DecBit* s,
                            int32_t DCF77_BitCnt,
                            int32_t DCF77_BitVal)
{
  // Reset?
  if(DCF77_BitCnt!=s->_PrevBitCnt+1)
  {
    s->ValidCnt = 0;
    s->_BitCntOk = -1;
  }
  s->_PrevBitCnt = DCF77_BitCnt;
  s->_DCFBits0 >>= 1;
  if((s->_DCFBits1&1)!=0) s->_DCFBits0|=1<<21; // 21 Bits "Start"
  s->_DCFBits1 >>= 1;
  if((s->_DCFBits2&1)!=0) s->_DCFBits1|=1<<14; // 15 Bits "Zeit"
  s->_DCFBits2 >>= 1;
  if(DCF77_BitVal!=0) s->_DCFBits2|=1<<23; // 23 Bits "Datum"
  s->Time_Second++;

  // Start
  if((s->_DCFBits0&0x0fff)!=0x07fe)
  {
    if(s->Time_Second>=60)
    {
      s->ValidCnt = 0;
      s->Time_Second = 0;
    }
    return;
  }

  bool Mod = false; // nderung zum vorherigen Datensatz?
  bool Err = false; // erkennbare Fehler im aktuellen DCF-Datensatz?

  // Zone
  {
    int Z = (s->_DCFBits0>>18)&1;
    if(((s->_DCFBits0>>19)&1)==(uint32_t)Z) Err=true;
    if(Z!=s->Zone_MEZS) Mod=true;
    s->Zone_MEZS = Z;
  }

  // Sekunden
  {
    s->Time_Second = 0;
  }

  // Minuten
  {
    int M0 = (s->_DCFBits1>> 0)&15;
    int M1 = (s->_DCFBits1>> 4)& 7;
    int M = 10*M1+M0;
    if(M0>9) Err=true;
    if(M>59) Err=true;
    if(IsOddParity((s->_DCFBits1>>0)&((1<<8)-1))) Err=true;
    s->Time_Minute = s->Time_Minute==59 ? 0 : s->Time_Minute+1;
    if(M!=s->Time_Minute) Mod=true;
    s->Time_Minute = M;
  }
  // Stunden
  {
    int H0 = (s->_DCFBits1>> 8)&15;
    int H1 = (s->_DCFBits1>>12)& 3;
    int H = 10*H1+H0;
    if(H0>9) Err=true;
    if(H>23) Err=true;
    if(IsOddParity((s->_DCFBits1>>8)&((1<<7)-1))) Err=true;
    if(s->Time_Minute==0)
    {
      s->Time_Hour = s->Time_Hour==23 ? 0 : s->Time_Hour+1;
    }
    if(H!=s->Time_Hour) Mod=true;
    s->Time_Hour = H;
  }
  // Datum
  {
    int d0 = (s->_DCFBits2>>(36-36))&15;
    int d1 = (s->_DCFBits2>>(40-36))& 3;
    int wd = (s->_DCFBits2>>(42-36))& 7;
    int m0 = (s->_DCFBits2>>(45-36))&15;
    int m1 = (s->_DCFBits2>>(49-36))& 1;
    int y0 = (s->_DCFBits2>>(50-36))&15;
    int y1 = (s->_DCFBits2>>(54-36))&15;
    int d = 10*d1+d0;
    int m = 10*m1+m0;
    int y = 10*y1+y0;
    if(d0>9) Err=true;
    if(m0>9) Err=true;
    if(y0>9) Err=true;
    if(y1>9) Err=true;
    if(wd<1) Err=true;
    if(d<1 || d>31) Err=true;
    if(m<1 || m>12) Err=true;
    // Wochentag korrekt?
    {
      // Wochentage im 21. Jahrhundet (d:1..31;  m:1..12; y:0..99)
      // mw[m-1] = int(2.6*((i+9)%12+1)-0.2)%7 ("Wikipedia Wochentagsberechnung")
      static const uint8_t mw[12] = { 0, 3, 2, 5, 0, 3, 5, 1, 4, 6, 2, 4 };
      int yy = m<3 ? y-1 : y;
      int w = ((d+mw[m-1]+yy+(yy>>2)+6)%7)+1;
      if(w!=wd) Err=true;
    }
    if(IsOddParity((s->_DCFBits2>>0)&((1<<23)-1))) Err=true;
    if(d !=s->Date_Day    ) Mod=true;
    if(wd!=s->Date_WeekDay) Mod=true;
    if(m !=s->Date_Month  ) Mod=true;
    if(y !=s->Date_Year   ) Mod=true;
    s->Date_Day     = d;
    s->Date_WeekDay = wd;
    s->Date_Month   = m;
    s->Date_Year    = y;
  }
  s->_BitCntOk = Err ? -1 : DCF77_BitCnt+60;
  if(Mod || Err) s->ValidCnt=0;
  if(!Err) s->ValidCnt++;
}

//############################################################## DCF77DecBit_Init
void DCF77DecBit_Init(TDCF77DecBit* s)
{
  for(int i=0; i<sizeof(*s); i++)
  {
    ((uint8_t*)s)[i] = 0;
  }
}

//############################################################## DCF77DecBit_Deinit
void DCF77DecBit_Deinit(TDCF77DecBit* s)
{
}
