
#include "stm32f10x.h"

//
// - LCResonanceMeter - misst LC-Resonanzfrequenz
//   Hilfsprogramm um einen DCF77-Schwingkreis zu prfen.
// - Schaltplan
//   LC-Parallelschwingkreis an GND und ber 1 kOhm Widerstand an PA6.
//   Um auch die Resonanzverschiebung des Verstrkers zu bercksichtigen,
//   kann dieser inkl. Stromversorgung angeschlossen bleiben.
// - Achtung!
//   Bei Anschluss einer Spule ohne Widerstand und Kondensator knnte der
//   Controller beschdigt werden!
// - Messergebnis
//   - das Ergebnis wird an USART1 (9600/8-N-1 an GPIO_PA9) ausgegeben
//   - alternativ Breakpoint auf "f_in_Hertz = f_in_Hertz;" setzen und
//     von Debugger anzeigen lassen
// - Technische Daten
//   - 3kHz - 570kHz
//   - 5% vermutete Genauigkeit
//   - etwa 100kOhm Eigendmpfung (=1/fADC/CADC=1/1.14MHz/8pF)
// - Funktionsweise
//   Frequenzmessung per DFT des Nachschwingens nach Impuls an LC.
//   Genau genommen wird die Abklingfrequenz gemessen. Diese ist je nach
//   Bandbreite mehr (161Hz Abweichung bei 10kHz Bandbreite) oder weniger
//   (1.61Hz bei 1kHz) kleiner als die Resonanzfrequenz. Die sich daraus
//   ergebene Falschabstimmung beim DCF77-Schwingkreis hat aber kaum
//   Auswirkung auf die Signalhhe (-0.05% bei 10000Hz Bandbreite).
//
// Copyright: Public Domain
//
// C.B., 2021
//

enum
{
  ADCCnt   = 1024,
  usPulse  = 1000,  // 1..1000
  ADCCyles = 14*4,  // (1.5 Sample&Hold + 12.5 Convert) * 4 PCLK2-Prescaler
  nAvrLoop =  200,  // Messdurchgnge je Messung
};

//################################################### DoADCSample
static void __attribute__((optimize("-Os"))) DoADCSample(int32_t* ADCArr)
{
  int16_t TmpArr[64];
  const int TmpCnt = sizeof(TmpArr)/sizeof(*TmpArr);

  // DMA
  RCC->AHBENR |= RCC_AHBENR_DMA1EN;
  DMA1_Channel1->CPAR  = (uint32_t)&ADC1->DR;
  DMA1_Channel1->CMAR  = (uint32_t)&TmpArr;
  DMA1_Channel1->CNDTR = TmpCnt;
  DMA1_Channel1->CCR = (DMA_CCR1_MEM2MEM *0) |
                       (DMA_CCR1_PL_0    *0) |
                       (DMA_CCR1_MSIZE_0 *1) |
                       (DMA_CCR1_PSIZE_0 *1) |
                       (DMA_CCR1_MINC    *1) |
                       (DMA_CCR1_PINC    *0) |
                       (DMA_CCR1_CIRC    *1) |
                       (DMA_CCR1_DIR     *0) |
                       (DMA_CCR1_TEIE    *0) |
                       (DMA_CCR1_HTIE    *0) |
                       (DMA_CCR1_TCIE    *0) |
                       (DMA_CCR1_EN      *1);

  // ADC
  RCC->CFGR   |= RCC_CFGR_ADCPRE_0*1; // 1 -> PCLK2/4
  RCC->APB2ENR|= RCC_APB2ENR_ADC1EN;
  {
    // Calibrate ADC
    ADC1->CR2   |= ADC_CR2_ADON;        // ADC ON (0->1)
    ADC1->CR2   |= ADC_CR2_RSTCAL;      // reset calibration
    while((ADC1->CR2&ADC_CR2_RSTCAL));  // wait for reset
    ADC1->CR2   |= ADC_CR2_CAL;         // start calibration
    while((ADC1->CR2&ADC_CR2_CAL)!=0);  // wait till calibration is done
    ADC1->CR2   &= ~ADC_CR2_ADON;       // ADC OFF
  }
  ADC1->CR2 |= ADC_CR2_ADON;
  ADC1->CR2 |= ADC_CR2_DMA;
  ADC1->SQR3  |= ADC_SQR3_SQ1_0*6;    // 6 -> PA6
  ADC1->SMPR2 |= ADC_SMPR2_SMP6_0*0;  // 0 -> 14 Takte (T0+Tx=12.5+1.5) fr PA6
  // ADC1->SMPR2 |= ADC_SMPR2_SMP6_0*4;  // 4 -> 54 Takte (T0+Tx=12.5+41.5) fr PA6
  ADC1->CR2 |= ADC_CR2_CONT;

  RCC->APB1ENR |= RCC_APB1ENR_TIM3EN;
  TIM3->CR1 |= TIM_CR1_CEN;
  TIM3->CNT = 0;

  // 3.3V Impuls ber 1k an Schwingkreis per PA6
  RCC->APB2ENR |= RCC_APB2ENR_IOPAEN;
  GPIOA->BSRR = 1<<6;
  GPIOA->CRL = (GPIOA->CRL&0xf0ffffff)|0x03000000; // Output PushPull
  while(TIM3->CNT<usPulse*64)
  {
  }
  // PA6 auf Floating
  GPIOA->CRL = (GPIOA->CRL&0xf0ffffff)|0x00000000; // Input Analog

  // ADC starten
  ADC1->CR2   |= ADC_CR2_ADON; // ADC Start
  DMA1->IFCR = DMA_IFCR_CHTIF1 | DMA_IFCR_CTCIF1;
  for(int iLoop=0; iLoop<ADCCnt; iLoop+=TmpCnt/2)
  {
    uint32_t Flag = (iLoop&(TmpCnt/2))==0 ? DMA_ISR_HTIF1 : DMA_ISR_TCIF1;
    while((DMA1->ISR&Flag)==0)
    {
    }
    DMA1->IFCR = (iLoop&(TmpCnt/2))==0 ? DMA_IFCR_CHTIF1 : DMA_IFCR_CTCIF1;
    int16_t* q = (iLoop&(TmpCnt/2))==0 ? TmpArr : TmpArr+TmpCnt/2;
    for(int i=0; i<TmpCnt/2; i++)
    {
      ADCArr[iLoop+i] += q[i];
    }
  }
  ADC1->CR2 = 0;
  DMA1_Channel1->CCR = 0;
}

//################################################### DoDFT
static int __attribute__((optimize("-Os"))) DoDFT(int32_t* ADCArr)
{
  const int nSinTab = 256;
  static const int8_t SinTab[256+64] =
  {
    // SinTab[i] = int(sin(2*M_PI*i/nSinTab)*127)
     0,    3,    6,    9,   12,   15,   18,   21,   24,   27,   30,   33,   36,   39,   42,   45,
    48,   51,   54,   57,   59,   62,   65,   67,   70,   73,   75,   78,   80,   82,   85,   87,
    89,   91,   94,   96,   98,  100,  102,  103,  105,  107,  108,  110,  112,  113,  114,  116,
   117,  118,  119,  120,  121,  122,  123,  123,  124,  125,  125,  126,  126,  126,  126,  126,
   127,  126,  126,  126,  126,  126,  125,  125,  124,  123,  123,  122,  121,  120,  119,  118,
   117,  116,  114,  113,  112,  110,  108,  107,  105,  103,  102,  100,   98,   96,   94,   91,
    89,   87,   85,   82,   80,   78,   75,   73,   70,   67,   65,   62,   59,   57,   54,   51,
    48,   45,   42,   39,   36,   33,   30,   27,   24,   21,   18,   15,   12,    9,    6,    3,
     0,   -3,   -6,   -9,  -12,  -15,  -18,  -21,  -24,  -27,  -30,  -33,  -36,  -39,  -42,  -45,
   -48,  -51,  -54,  -57,  -59,  -62,  -65,  -67,  -70,  -73,  -75,  -78,  -80,  -82,  -85,  -87,
   -89,  -91,  -94,  -96,  -98, -100, -102, -103, -105, -107, -108, -110, -112, -113, -114, -116,
  -117, -118, -119, -120, -121, -122, -123, -123, -124, -125, -125, -126, -126, -126, -126, -126,
  -127, -126, -126, -126, -126, -126, -125, -125, -124, -123, -123, -122, -121, -120, -119, -118,
  -117, -116, -114, -113, -112, -110, -108, -107, -105, -103, -102, -100,  -98,  -96,  -94,  -91,
   -89,  -87,  -85,  -82,  -80,  -78,  -75,  -73,  -70,  -67,  -65,  -62,  -59,  -57,  -54,  -51,
   -48,  -45,  -42,  -39,  -36,  -33,  -30,  -27,  -24,  -21,  -18,  -15,  -12,   -9,   -6,   -3,
     0,    3,    6,    9,   12,   15,   18,   21,   24,   27,   30,   33,   36,   39,   42,   45,
    48,   51,   54,   57,   59,   62,   65,   67,   70,   73,   75,   78,   80,   82,   85,   87,
    89,   91,   94,   96,   98,  100,  102,  103,  105,  107,  108,  110,  112,  113,  114,  116,
   117,  118,  119,  120,  121,  122,  123,  123,  124,  125,  125,  126,  126,  126,  126,  126,
  };
  // Daten fr DFT anpassen
  {
    // 1. Mittelwert entfernen
    // 2. Symmetrie herstellen
    //    (der ADC-Wandler schneidet neg. Halbwelle fast vollstndig ab -> strt
    //    im Spektrum bei kleinen Frequenzen -> positive Halbwelle begrenzen)
    // 3. erste Halbwelle lschen
    int Sum = 0;
    for(int i=3*ADCCnt/4; i<ADCCnt; i++)
    {
      Sum += ADCArr[i];
    }
    int Avr = Sum/(ADCCnt/4);
    for(int i=0; i<ADCCnt; i++)
    {
      ADCArr[i] -= Avr;
      if(ADCArr[i]>Avr) ADCArr[i]=Avr;
    }
    for(int i=0; i<ADCCnt && ADCArr[i]<0; i++)
    {
      ADCArr[i] = 0;
    }
  }
  int  wMax = 0;
  int ScaleUp = 1;
  for(int iLoop=0; iLoop<2; iLoop++)
  {
    int64_t aMax =0;
    int w0 = iLoop==0 ?        1 : (wMax-2)*ScaleUp;
    int w1 = iLoop==0 ? ADCCnt/2 : (wMax+2)*ScaleUp;
    if(w0<     1*ScaleUp) w0=     1*ScaleUp;
    if(w1>ADCCnt*ScaleUp) w1=ADCCnt*ScaleUp;
    for(int w=w0; w<w1; w++)
    {
      int64_t sSum = 0;
      int64_t cSum = 0;
      for(int i=0; i<ADCCnt; i++)
      {
        int p = (nSinTab*w*i/ScaleUp/ADCCnt)&(nSinTab-1);
        sSum += ADCArr[i]*SinTab[p+0*nSinTab/4];
        cSum += ADCArr[i]*SinTab[p+1*nSinTab/4];
      }
      int64_t a = sSum*sSum+cSum*cSum;
      if(a>aMax)
      {
        aMax = a;
        wMax = w;
      }
    }
    ScaleUp = 32;
  }
  return (int32_t)((int64_t)64000000*wMax/(ADCCyles*ADCCnt*ScaleUp));
}

//################################################### USART_out
static void USART_out(char Char)
{
  // TXD-Ausgang ist GPIO_PA9
  const int Baud = 9600;
  const int fMHz = 64;
  if((USART1->CR1&USART_CR1_UE)==0)
  {
    RCC->APB2ENR |= RCC_APB2ENR_IOPAEN;
    GPIOA->CRH = (GPIOA->CRH&0xffffff0f)|0x000000a0;
    RCC->APB2ENR |= RCC_APB2ENR_USART1EN;
    USART1->CR1 = USART_CR1_UE;
    int divval = (fMHz/1*1000000+Baud/2)/Baud; // 16..65535
    USART1->BRR = divval;
    USART1->CR1 |= USART_CR1_TE;
  }
  USART1->DR = Char;
  while((USART1->SR&USART_SR_TXE)==0)
  {
  }
}

//################################################### main_LCResonanceMeter
int main_LCResonanceMeter(void)
{
  int32_t ADCArr[ADCCnt];

  // HSE auf 64 MHz (Q&D)
  RCC->CFGR = 0;
  RCC->CR &= ~(RCC_CR_HSEON|RCC_CR_PLLON);
  RCC->CR |= RCC_CR_HSEON;
  while((RCC->CR|RCC_CR_HSERDY)==0);
  RCC->CFGR |= RCC_CFGR_PLLSRC;
  RCC->CFGR |= RCC_CFGR_SW_0*1;
  while((RCC->CFGR&RCC_CFGR_SWS)!=RCC_CFGR_SWS_0*1);
  RCC->CFGR |= RCC_CFGR_PLLMULL8; // 8*(8MHz/1)
  FLASH->ACR = FLASH_ACR_LATENCY_1*2;
  RCC->CR |= RCC_CR_PLLON;
  RCC->CFGR = (RCC->CFGR&~RCC_CFGR_SW)|(RCC_CFGR_SW_0*2);

  while(1)
  {
    for(int i=0; i<ADCCnt; i++)
    {
      ADCArr[i] = 0;
    }
    for(int i=0; i<nAvrLoop; i++)
    {
      DoADCSample(ADCArr);
    }
    int f_in_Hertz = DoDFT(ADCArr);
    f_in_Hertz = f_in_Hertz;

    // USART-Ausgabe von "f_in_Hertz"
    {
      int v = f_in_Hertz;
      int cv = 1000000000;
      for(; cv>v && cv>1; cv/=10)
      {
      }
      for(; cv>0; cv/=10)
      {
        int n = v/cv;
        USART_out(n+48);
        v -= n*cv;
      }
      USART_out(' ');
      USART_out('H');
      USART_out('z');
      USART_out('\r');
      USART_out('\n');
    }
  }
  return 0;
}
